# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import pytest

import xtrack as xt
import xcoll as xc
from xcoll.accessors import XcollAccessor



import pytest
from accessors import XcollAccessor

class InnerClass:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
    def get(self, kk):
        try:
            return getattr(self, kk)
        except AttributeError:
            raise KeyError
    def __str__(self): # For test debugging
        return f'InnerClass({self.__dict__})'

class OuterClass:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, f'_{k}', v)
    def get(self, kk):
        try:
            return getattr(self, f'_{kk}')
        except AttributeError:
            raise KeyError
    def __str__(self): # For test debugging
        return f'OuterClass({self.__dict__})'

outer_is_instance = True
inner_is_instance = True
extra_elements = True

@pytest.mark.parametrize("outer_is_instance", [False, True
                        ], ids=["outer_dict", "outer_instance"])
@pytest.mark.parametrize("inner_is_instance", [False, True
                        ], ids=["inner_dict", "inner_instance"])
@pytest.mark.parametrize("extra_elements", [False, True
                        ], ids=["no_extra_elements", "with_extra_elements"])
def test_accessor(outer_is_instance, inner_is_instance, extra_elements):
    # Create different combinations of db types:
    inner = InnerClass if inner_is_instance else dict
    outer = OuterClass if outer_is_instance else dict
    el1 = inner(a=1, b=2, d=10)
    el2 = inner(a=3, b=4, d=10)
    el3 = inner(a=5, c=6, d=10)
    el4 = inner(b=7, e=23)
    if extra_elements:
        db = outer(el1=el1, el2=el2, el3=el3, el4=el4)
    else:
        db = outer(el1=el1, el2=el2, el3=el3)

    # Create the accessor:
    if outer_is_instance:
        _dbtype = 'test class'
        acc = XcollAccessor(db=db, names=['el1', 'el2', 'el3'], _dbtype=_dbtype, _typename='test element')
    else:
        _dbtype = 'test dict'
        if extra_elements:
            acc = XcollAccessor(db=db, names=['el1', 'el2', 'el3'], _dbtype=_dbtype, _typename='test element')
        else:
            acc = XcollAccessor(db=db, _dbtype=_dbtype, _typename='test element')

    # The accessor should behave like a dict:
    assert len(acc) == 3
    assert set(acc.names) == {'el1', 'el2', 'el3'}
    assert set(acc.keys()) == {'el1', 'el2', 'el3'}
    assert all(isinstance(val, inner) for val in acc.values())
    assert all(isinstance(val, inner) for _, val in acc.items())
    assert {nn for nn, _ in acc.items()} == {'el1', 'el2', 'el3'}
    if not inner_is_instance:
        assert list(acc.values()) == [{'a': 1, 'b': 2, 'd': 10},
                                      {'a': 3, 'b': 4, 'd': 10},
                                      {'a': 5, 'c': 6, 'd': 10}]
        assert list(acc.items()) == [('el1', {'a': 1, 'b': 2, 'd': 10}),
                                     ('el2', {'a': 3, 'b': 4, 'd': 10}),
                                     ('el3', {'a': 5, 'c': 6, 'd': 10})]
    assert acc['el1'] is db.get('el1')
    assert acc['el2'] is db.get('el2')
    assert acc['el3'] is db.get('el3')
    if inner_is_instance:
        assert {nn: acc['el1'].get(nn) for nn in acc['el1'].__dict__.keys()} == {'a': 1, 'b': 2, 'd': 10}
        assert {nn: acc['el2'].get(nn) for nn in acc['el2'].__dict__.keys()} == {'a': 3, 'b': 4, 'd': 10}
        assert {nn: acc['el3'].get(nn) for nn in acc['el3'].__dict__.keys()} == {'a': 5, 'c': 6, 'd': 10}
    else:
        assert acc['el1'] == {'a': 1, 'b': 2, 'd': 10}
        assert acc['el2'] == {'a': 3, 'b': 4, 'd': 10}
        assert acc['el3'] == {'a': 5, 'c': 6, 'd': 10}
    assert 'el1' in acc
    assert 'el2' in acc
    assert 'el3' in acc

    # Accessing a non-existing element or elements that are not represented
    # (i.e. not in names) should raise an error:
    assert 'el4' not in acc
    assert 'non_existing' not in acc
    with pytest.raises(KeyError) as err:
        acc['el4']
    assert str(err.value) == f"'Test element `el4` not found in {_dbtype}!'"
    if extra_elements:
        assert isinstance(db.get('el4'), inner) # Should not have been modified
    else:
        with pytest.raises(KeyError):           # Should not have been added
            _ = db['el4'] if outer == dict else db.get('el4')
    with pytest.raises(KeyError) as err:
        acc['non_existing']
    assert str(err.value) == f"'Test element `non_existing` not found in {_dbtype}!'"
    with pytest.raises(KeyError):  # Should not have been added
        _ = db['non_existing'] if outer == dict else db.get('non_existing')

    # Iteration over the accessor should yield the elements (not like a dict, which yields keys):
    for i, nn in enumerate(acc):
        assert nn == db.get(f'el{i+1}')

    # Accessing an attribute should return a dict of element -> attribute value, omitting
    # elements that do not have the attribute:
    assert acc.a == {'el1': 1, 'el2': 3, 'el3': 5}
    assert acc.b == {'el1': 2, 'el2': 4}
    assert acc.c == {'el3': 6}
    with pytest.raises(AttributeError) as err:
        _ = acc.e
    assert str(err.value) == f"Attribute `e` not found in {_dbtype}!"

    # If all elements have the same value for an attribute, return that single value:
    assert acc.d == 10

    # Setting an attribute with a dict should set the attribute for the specified elements:
    acc.a = {'el1': 10, 'el2': 30}
    assert acc.a == {'el1': 10, 'el2': 30, 'el3': 5}
    # And the attribute a in the original db should have been updated:
    assert db.get('el1').get('a') == 10
    assert db.get('el2').get('a') == 30
    assert db.get('el3').get('a') == 5
    # But not the other attributes:
    assert db.get('el1').get('b') == 2
    assert db.get('el2').get('b') == 4
    if extra_elements:
        assert db.get('el4').get('b') == 7
    assert db.get('el3').get('c') == 6
    assert db.get('el1').get('d') == 10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    if extra_elements:
        assert db.get('el4').get('e') == 23

    # Setting an attribute with a single value should set it for all elements that have that attribute:
    acc.b = 5000
    assert acc.b == {'el1': 5000, 'el2': 5000}
    # And the attribute b in the original db should have been updated:
    assert db.get('el1').get('b') == 5000
    assert db.get('el2').get('b') == 5000
    # But not the other attributes nor b in elements that are not represented by the accessor:
    assert not hasattr(db.get('el3'), 'b')
    if extra_elements:
        assert db.get('el4').get('b') == 7
    assert db.get('el1').get('a') == 10
    assert db.get('el2').get('a') == 30
    assert db.get('el3').get('a') == 5
    assert db.get('el3').get('c') == 6
    assert db.get('el1').get('d') == 10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    if extra_elements:
        assert db.get('el4').get('e') == 23

    # Testing invalid attribute setting:
    with pytest.raises(AttributeError) as err:
        acc.b = {'el1': 5000, 'el3': 9000}
    assert str(err.value) == "Attribute `b` not found in test element `el3`!"
    with pytest.raises(AttributeError) as err:
        acc.c = {'el1': -800}
    assert str(err.value) == "Attribute `c` not found in test element `el1`!"
    assert acc.c == {'el3': 6}
    with pytest.raises(KeyError) as err:
        acc.c = {'el7': -800}
    assert str(err.value) == f"'Test element `el7` not found in {_dbtype}!'"
    assert acc.c == {'el3': 6}
    with pytest.raises(AttributeError) as err:
        acc.e = 80
    assert str(err.value) == f"Attribute `e` not found in {_dbtype}!"
    # The db should not have been modified:
    assert db.get('el1').get('a') == 10
    assert db.get('el2').get('a') == 30
    assert db.get('el3').get('a') == 5
    assert db.get('el1').get('b') == 5000
    assert db.get('el2').get('b') == 5000
    if extra_elements:
        assert db.get('el4').get('b') == 7
    assert db.get('el3').get('c') == 6
    assert db.get('el1').get('d') == 10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    if extra_elements:
        assert db.get('el4').get('e') == 23

    # Setting attributes by item assignment (non-specified attributes should
    # be left unchanged, and new attributes should be added):
    acc['el1'] = {'a': -10, 'd': -10, 'f': -999}
    assert db.get('el1').get('a') == -10
    assert db.get('el2').get('a') == 30
    assert db.get('el3').get('a') == 5
    assert db.get('el1').get('b') == 5000
    assert db.get('el2').get('b') == 5000
    if extra_elements:
        assert db.get('el4').get('b') == 7
    assert db.get('el3').get('c') == 6
    assert db.get('el1').get('d') == -10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    if extra_elements:
        assert db.get('el4').get('e') == 23
    assert db.get('el1').get('f') == -999

    # Setting attributes by item assignment can only be done with a dict:
    with pytest.raises(ValueError) as err:
        acc['el1'] = 8
    assert str(err.value) == "Can only set test element `el1` to a settings dict!"
    assert db.get('el1').get('a') == -10
    assert db.get('el2').get('a') == 30
    assert db.get('el3').get('a') == 5
    assert db.get('el1').get('b') == 5000
    assert db.get('el2').get('b') == 5000
    if extra_elements:
        assert db.get('el4').get('b') == 7
    assert db.get('el3').get('c') == 6
    assert db.get('el1').get('d') == -10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    if extra_elements:
        assert db.get('el4').get('e') == 23
    assert db.get('el1').get('f') == -999


@pytest.mark.parametrize("outer_is_instance", [False, True
                        ], ids=["outer_dict", "outer_instance"])
@pytest.mark.parametrize("inner_is_instance", [False, True
                        ], ids=["inner_dict", "inner_instance"])
@pytest.mark.parametrize("extra_elements", [False, True
                        ], ids=["no_extra_elements", "with_extra_elements"])
def test_accessor_family(outer_is_instance, inner_is_instance, extra_elements):
    # Create different combinations of db types:
    inner = InnerClass if inner_is_instance else dict
    outer = OuterClass if outer_is_instance else dict
    el1 = inner(a=1, b=2, d=10, family='A')
    el2 = inner(a=3, b=4, d=10, family='B')
    el3 = inner(a=5, b=19, c=6, d=10, family='A', non_family_attributes=['b'])
    el4 = inner(a=9, d=10)
    el5 = inner(b=7, e=23, family='B')
    if extra_elements:
        db = outer(el1=el1, el2=el2, el3=el3, el4=el4, el5=el5)
    else:
        db = outer(el1=el1, el2=el2, el3=el3, el4=el4)

    # Create the accessor:
    if outer_is_instance:
        _dbtype = 'test class'
        acc = XcollAccessor(db=db, names=['el1', 'el2', 'el3', 'el4'],
                            _dbtype=_dbtype, _typename='test element')
    else:
        _dbtype = 'test dict'
        if extra_elements:
            acc = XcollAccessor(db=db, names=['el1', 'el2', 'el3', 'el4'],
                                _dbtype=_dbtype, _typename='test element')
        else:
            acc = XcollAccessor(db=db, _dbtype=_dbtype, _typename='test element')

    # Check the families
    assert acc.family_names == ['A', 'B']
    assert acc.families == {'A': ['el1', 'el3'], 'B': ['el2']}
    assert isinstance(acc['A'], XcollAccessor)
    assert isinstance(acc['B'], XcollAccessor)
    assert acc['A']._is_family_sub_accessor
    assert acc['B']._is_family_sub_accessor
    assert set(acc['A'].names) == {'el1', 'el3'}
    assert set(acc['B'].names) == {'el2'}

    # Get family attributes:
    assert acc['A'].a == {'el1': 1, 'el3': 5}
    assert acc['A'].b == {'el1': 2, 'el3': 19}
    assert acc['A'].d == 10
    assert acc['B'].a == 3
    assert acc['B'].b == 4
    assert acc['B'].d == 10

    # Attribute c is not a family attribute of A, so accessing it should raise an error:
    with pytest.raises(AttributeError) as err:
        _ = acc['A'].c
    assert str(err.value) == f"Attribute `c` is not a family attribute, as test " \
                           + f"element `el1` does not have it! If this is intentional, " \
                           + f"add the attribute to `el1`s non_family_attributes list!"

    # Set family attribute a (should update el1 and el3):
    acc['A'].a = -1
    assert acc['A'].a == -1
    # Check the db is updated correctly:
    assert db.get('el1').get('a') == -1
    assert db.get('el2').get('a') == 3
    assert db.get('el3').get('a') == -1
    assert db.get('el4').get('a') == 9
    assert db.get('el1').get('b') == 2
    assert db.get('el2').get('b') == 4
    assert db.get('el3').get('b') == 19
    if extra_elements:
        assert db.get('el5').get('b') == 7
    assert db.get('el3').get('c') == 6
    assert db.get('el1').get('d') == 10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    assert db.get('el4').get('d') == 10
    if extra_elements:
        assert db.get('el5').get('e') == 23

    # Set family attribute b (should only update el1, as el3 has b as non_family_attribute):
    acc['A'].b = -2
    assert acc['A'].b == {'el1': -2, 'el3': 19}
    # Check the db is updated correctly:
    assert db.get('el1').get('a') == -1
    assert db.get('el2').get('a') == 3
    assert db.get('el3').get('a') == -1
    assert db.get('el4').get('a') == 9
    assert db.get('el1').get('b') == -2
    assert db.get('el2').get('b') == 4
    assert db.get('el3').get('b') == 19
    if extra_elements:
        assert db.get('el5').get('b') == 7
    assert db.get('el3').get('c') == 6
    assert db.get('el1').get('d') == 10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    assert db.get('el4').get('d') == 10
    if extra_elements:
        assert db.get('el5').get('e') == 23

    # Setting family attributes by item assignment:
    acc['A'] = {'a': -1000, 'b': -5000, 'c': -200, 'f': -999}
    # Check the db is updated correctly:
    assert db.get('el1').get('a') == -1000
    assert db.get('el2').get('a') == 3
    assert db.get('el3').get('a') == -1000
    assert db.get('el4').get('a') == 9
    assert db.get('el1').get('b') == -5000
    assert db.get('el2').get('b') == 4
    assert db.get('el3').get('b') == 19
    if extra_elements:
        assert db.get('el5').get('b') == 7
    assert db.get('el1').get('c') == -200
    assert db.get('el3').get('c') == -200
    assert db.get('el1').get('d') == 10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    assert db.get('el4').get('d') == 10
    if extra_elements:
        assert db.get('el5').get('e') == 23
    assert db.get('el1').get('f') == -999
    assert db.get('el3').get('f') == -999

    # Setting attributes by item assignment can only be done with a dict:
    with pytest.raises(ValueError) as err:
        acc['A'] = 8
    assert str(err.value) == "Can only set family `A` to a settings dict!"
    # Check the db is unchanged:
    assert db.get('el1').get('a') == -1000
    assert db.get('el2').get('a') == 3
    assert db.get('el3').get('a') == -1000
    assert db.get('el4').get('a') == 9
    assert db.get('el1').get('b') == -5000
    assert db.get('el2').get('b') == 4
    assert db.get('el3').get('b') == 19
    if extra_elements:
        assert db.get('el5').get('b') == 7
    assert db.get('el1').get('c') == -200
    assert db.get('el3').get('c') == -200
    assert db.get('el1').get('d') == 10
    assert db.get('el2').get('d') == 10
    assert db.get('el3').get('d') == 10
    assert db.get('el4').get('d') == 10
    if extra_elements:
        assert db.get('el5').get('e') == 23
    assert db.get('el1').get('f') == -999
    assert db.get('el3').get('f') == -999


def test_accessor_family_inconsistency():
    el1 = {'a': 1, 'b': 2, 'd': 10, 'family': 'A'}
    el2 = {'a': 3, 'b': 4, 'd': 10, 'family': 'B'}
    el3 = {'a': 5, 'b': 19, 'c': 6, 'd': 10, 'family': 'A', 'non_family_attributes': ['b']}

    el2['family'] = 7
    with pytest.raises(ValueError) as err:
        acc = XcollAccessor(db={'el1': el1, 'el2': el2, 'el3': el3},
                            _dbtype='test dict', _typename='test element')
    assert str(err.value) == "Test element `el2` has a `family` attribute that is not a string!"

    el2['family'] = 'B'
    el3.pop('family')
    with pytest.raises(AttributeError) as err:
        acc = XcollAccessor(db={'el1': el1, 'el2': el2, 'el3': el3},
                            _dbtype='test dict', _typename='test element')
    assert str(err.value) == "Test element `el3` has `non_family_attributes` but no `family` attribute!"

    el3['family'] = 'A'
    el3['non_family_attributes'] = 'b'  # This is OK
    acc = XcollAccessor(db={'el1': el1, 'el2': el2, 'el3': el3},
                        _dbtype='test dict', _typename='test element')
    assert acc['el3']['non_family_attributes'] == ['b']
    assert el3['non_family_attributes'] == ['b']
    el3['non_family_attributes'] = 8  # This is not
    with pytest.raises(ValueError) as err:
        acc = XcollAccessor(db={'el1': el1, 'el2': el2, 'el3': el3},
                            _dbtype='test dict', _typename='test element')
    assert str(err.value) == "Test element `el3` has `non_family_attributes` but it is not iterable!"

    el3['non_family_attributes'] = ['f']
    with pytest.raises(AttributeError) as err:
        acc = XcollAccessor(db={'el1': el1, 'el2': el2, 'el3': el3},
                            _dbtype='test dict', _typename='test element')
    assert str(err.value) == "Test element `el3` has attribute `f` in `non_family_attributes` but it is not present!"

    el3['non_family_attributes'] = ['family']
    with pytest.raises(ValueError) as err:
        acc = XcollAccessor(db={'el1': el1, 'el2': el2, 'el3': el3},
                            _dbtype='test dict', _typename='test element')
    assert str(err.value) == "Test element `el3` has `family` in `non_family_attributes`!"

    el3['non_family_attributes'] = ['non_family_attributes']
    with pytest.raises(ValueError) as err:
        acc = XcollAccessor(db={'el1': el1, 'el2': el2, 'el3': el3},
                            _dbtype='test dict', _typename='test element')
    assert str(err.value) == "Test element `el3` has `non_family_attributes` in `non_family_attributes`!"



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from ..beam_elements import *


DEFAULT_XCOLL_ELEMENTS = [
    BlackAbsorber,
    BlackCrystal,
    TransparentCollimator,
    TransparentCrystal,
    EverestBlock,
    EverestCollimator,
    EverestCrystal,
    # FlukaCollimator,
    # FlukaCrystal,
    # Geant4Collimator,
    # Geant4Crystal,
    BlowUp,
    # ParticleStatsMonitor,
    EmittanceMonitor
]

EXTRA_XCOLL_ELEMENTS = []



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xobjects as xo
import xtrack as xt

from .base import BaseBlock, BaseCollimator, BaseCrystal, InvalidXcoll
from ..scattering_routines.everest import EverestEngine
from ..materials import Material, _DEFAULT_MATERIAL, _resolve_material
from ..general import _pkg_root


# TODO:
#      We want these elements to behave as if 'iscollective = True' when doing twiss etc (because they would ruin the CO),
#      but as if 'iscollective = False' for normal tracking as it is natively in C...
#      Currently this is achieved with the hack '_tracking' which defaults to False after installation in the line, and is
#      only activated around the track command. Furthermore, because of 'iscollective = False' we need to specify
#      get_backtrack_element. We want it nicer..


class EverestBlock(BaseBlock):
    _xofields = {**BaseBlock._xofields,
        '_material':        Material,
        'rutherford_rng':   xt.RandomRutherford,
        '_tracking':        xo.Int8
    }

    isthick = True
    needs_rng = True
    allow_track = True
    allow_double_sided = False
    behaves_like_drift = True
    allow_rot_and_shift = False
    allow_loss_refinement = True
    skip_in_loss_location_refinement = True

    _noexpr_fields         = {*BaseBlock._noexpr_fields, 'material'}
    _skip_in_to_dict       = [*BaseBlock._skip_in_to_dict, '_material', '_tracking']
    _store_in_to_dict      = [*BaseBlock._store_in_to_dict, 'material']
    _internal_record_class = BaseBlock._internal_record_class

    _depends_on = [BaseBlock, EverestEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','everest_block.h')
    ]

    _kernels = {
        'EverestBlock_set_material': xo.Kernel(
                c_name='EverestBlock_set_material',
                args=[xo.Arg(xo.ThisClass, name='el')]
            )
        }


    def __init__(self, **kwargs):
        to_assign = {}
        if '_xobject' not in kwargs:
            to_assign['material'] = kwargs.pop('material', None)
            kwargs['_material'] = _DEFAULT_MATERIAL
            kwargs.setdefault('rutherford_rng', xt.RandomRutherford())
            kwargs.setdefault('_tracking', True)
        super().__init__(**kwargs)
        for key, val in to_assign.items():
            setattr(self, key, val)


    @property
    def material(self):
        if self._material != _DEFAULT_MATERIAL:
            return self._material

    @material.setter
    def material(self, material):
        material = _resolve_material(material)
        if self.material != material:
            self._material = material
            self.EverestBlock_set_material(el=self)

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context,
                                 _buffer=_buffer, _offset=_offset)


class EverestCollimator(BaseCollimator):
    _xofields = {**BaseCollimator._xofields,
        '_material':        Material,
        'rutherford_rng':   xt.RandomRutherford,
        '_tracking':        xo.Int8
    }

    isthick = True
    needs_rng = True
    allow_track = True
    allow_double_sided = True
    behaves_like_drift = True
    allow_rot_and_shift = False
    allow_loss_refinement = True
    skip_in_loss_location_refinement = True

    _noexpr_fields         = {*BaseCollimator._noexpr_fields, 'material'}
    _skip_in_to_dict       = [*BaseCollimator._skip_in_to_dict, '_material']
    _store_in_to_dict      = [*BaseCollimator._store_in_to_dict, 'material']
    _internal_record_class = BaseCollimator._internal_record_class

    _depends_on = [BaseCollimator, EverestEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','everest_collimator.h')
    ]

    _kernels = {
        'EverestCollimator_set_material': xo.Kernel(
                c_name='EverestCollimator_set_material',
                args=[xo.Arg(xo.ThisClass, name='el')]
            )
        }


    def __init__(self, **kwargs):
        to_assign = {}
        if '_xobject' not in kwargs:
            to_assign['material'] = kwargs.pop('material', None)
            kwargs['_material'] = _DEFAULT_MATERIAL
            kwargs.setdefault('rutherford_rng', xt.RandomRutherford())
            kwargs.setdefault('_tracking', True)
        super().__init__(**kwargs)
        for key, val in to_assign.items():
            setattr(self, key, val)

    @property
    def material(self):
        if self._material != _DEFAULT_MATERIAL:
            return self._material

    @material.setter
    def material(self, material):
        material = _resolve_material(material)
        if self.material != material:
            self._material = material
            self.EverestCollimator_set_material(el=self)

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context,
                            _buffer=_buffer, _offset=_offset)


class EverestCrystal(BaseCrystal):
    _xofields = {**BaseCrystal._xofields,
        'miscut':             xo.Float64,
        '_orient':            xo.Int8,
        '_critical_angle':    xo.Float64,
        '_critical_radius':   xo.Float64,
        '_material':          Material,
        'rutherford_rng':     xt.RandomRutherford,
        '_tracking':          xo.Int8
    }

    isthick = True
    needs_rng = True
    allow_track = True
    allow_double_sided = False
    behaves_like_drift = True
    allow_rot_and_shift = False
    allow_loss_refinement = True
    skip_in_loss_location_refinement = True

    _noexpr_fields         = {*BaseCrystal._noexpr_fields, 'material', 'lattice'}
    _skip_in_to_dict       = [*BaseCrystal._skip_in_to_dict, '_orient', '_material']
    _store_in_to_dict      = [*BaseCrystal._store_in_to_dict, 'lattice', 'material']
    _internal_record_class = BaseCrystal._internal_record_class

    _depends_on = [BaseCrystal, EverestEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','everest_crystal.h')
    ]

    _kernels = {
        'EverestCrystal_set_material': xo.Kernel(
                c_name='EverestCrystal_set_material',
                args=[xo.Arg(xo.ThisClass, name='el')]
            )
        }


    def __init__(self, **kwargs):
        to_assign = {}
        if '_xobject' not in kwargs:
            to_assign['material'] = kwargs.pop('material', None)
            kwargs['_material'] = _DEFAULT_MATERIAL
            to_assign['lattice'] = kwargs.pop('lattice', 'strip')
            kwargs.setdefault('miscut', 0)
            kwargs.setdefault('rutherford_rng', xt.RandomRutherford())
            kwargs.setdefault('_tracking', True)
        super().__init__(**kwargs)
        for key, val in to_assign.items():
            setattr(self, key, val)


    @property
    def material(self):
        if self._material != _DEFAULT_MATERIAL:
            return self._material

    @material.setter
    def material(self, material):
        material = _resolve_material(material, everest_crystal=True)
        if self.material != material:
            self._material = material
            self.EverestCrystal_set_material(el=self)

    @property
    def critical_angle(self):
        return self._critical_angle if abs(self._critical_angle) > 1.e-12 else None

    @property
    def critical_radius(self):
        return self._critical_radius if abs(self._critical_radius) > 1.e-10 else None

    @property
    def lattice(self):
        if self._orient == 1:
            return 'strip'
        elif self._orient == 2:
            return 'quasi-mosaic'
        else:
            raise ValueError(f"Illegal value {self._orient} for '_orient'!")

    @lattice.setter
    def lattice(self, lattice):
        if lattice == 'strip' or lattice == '110' or lattice == 110:
            self._orient = 1
        elif lattice == 'quasi-mosaic' or lattice == '111' or lattice == 111:
            self._orient = 2
        else:
            raise ValueError(f"Illegal value {lattice} for 'lattice'! "
                            + "Only use 'strip' (110) or 'quasi-mosaic' (111).")


    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context,
                            _buffer=_buffer, _offset=_offset)





# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from .base import BaseBlock, BaseCollimator, BaseCrystal
from .absorber import BlackAbsorber, BlackCrystal
from .transparent import TransparentCollimator, TransparentCrystal
from .everest import EverestBlock, EverestCollimator, EverestCrystal
from .fluka import FlukaCollimator, FlukaCrystal
from .geant4 import Geant4Collimator, Geant4CollimatorTip, Geant4Crystal
from .blowup import BlowUp
from .monitor import ParticleStatsMonitor, EmittanceMonitor

block_classes = tuple(v for v in globals().values()
                      if isinstance(v, type) and issubclass(v, BaseBlock)
                      and v != BaseBlock and v != BaseCollimator
                      and v != BaseCrystal)
# Includes crystals
collimator_classes = tuple(v for v in globals().values() if isinstance(v, type)
                           and (issubclass(v, BaseCollimator)
                                or issubclass(v, BaseCrystal))
                           and v != BaseCollimator and v != BaseCrystal)
crystal_classes = tuple(v for v in globals().values() if isinstance(v, type)
                        and issubclass(v, BaseCrystal) and v != BaseCrystal)

element_classes = block_classes + (BlowUp, ParticleStatsMonitor, EmittanceMonitor)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from numbers import Number


_NORMS = ["total", "coll_max", "max", "none", "raw", "deposited_energy", "deposited_energy_per_length"]


def plot_lossmap(lossmap: dict, *,
                 ax=None,
                 show=True,
                 zoom=None,
                 savefig=None,
                 titles=None,
                 norm="total",
                 energy=False,
                 rescaled_beam_intensity=None,
                 cold_regions=None,
                 warm_regions=None,
                 **kwargs):
    import matplotlib.pyplot as plt

    # Resolve zoom and titles parameters
    zoom = _resolve_zoom(zoom)
    if titles is None:
        titles = []
    if not hasattr(titles, "__iter__") or isinstance(titles, str):
        titles = [titles]
    if zoom and titles:
        if len(titles) != len(zoom) + 1:
            raise ValueError("When zoom is used, titles must be a list of as many titles as zoom intervals plus one.")

    font = {'family': 'serif', 'size': 17}
    format_dict = {f"font.{prop}": font[prop] for prop in font}
    with plt.rc_context(format_dict):
        if ax is None:
            _, ax = plt.subplots(len(zoom) + 1, 1, figsize=(16, (len(zoom) + 1)*4))
        if not zoom:
            ax = np.array([ax])
        fig = ax[0].figure
        if not isinstance(ax, np.ndarray) or ax.ndim != 1 or ax.shape[0] != len(zoom) + 1:
            raise ValueError("When zoom is used, ax must be a list of as many axes as zoom intervals.")
        xlim = kwargs.pop("xlim", None)
        _plot_lossmap_base(lossmap, ax=ax[0], norm=norm, energy=energy, xlim=xlim,
                           rescaled_beam_intensity=rescaled_beam_intensity, cold_regions=cold_regions,
                           warm_regions=warm_regions, **kwargs)
        if titles:
            ax[0].set_title(titles[0])
        for i, zz in enumerate(zoom):
            _plot_lossmap_base(lossmap, ax=ax[i+1], norm=norm, energy=energy, xlim=zz,
                               rescaled_beam_intensity=rescaled_beam_intensity, cold_regions=cold_regions,
                               warm_regions=warm_regions, **kwargs)
            if titles:
                ax[i+1].set_title(titles[i+1])

        plt.tight_layout()
        if savefig:
            plt.savefig(savefig, dpi=300, bbox_inches='tight')
        if show:
            plt.show()
        else:
            plt.close()
    if zoom:
        return fig, ax
    else:
        return fig, ax[0]


def _plot_lossmap_base(lossmap: dict, *,
                       ax=None,
                       grid=True,
                       norm="total",
                       normalise_by_length=None,
                       energy=False,
                       rescaled_beam_intensity=None,
                       cold_regions=None,
                       warm_regions=None,
                       xlim=None,
                       ylim=None,
                       x_label=None,
                       y_label=None,
                       xticks=None,
                       yticks=None,
                       legend=True):
    import matplotlib.pyplot as plt
    if not isinstance(ax, plt.Axes):
        raise ValueError("ax must be a matplotlib Axes instance.")

    # Resolve normalisation
    if isinstance(norm, str) and norm.lower() in _NORMS:
        norm = norm.lower()
    elif not isinstance(norm, Number):
        raise ValueError(f"Norm must be one of {_NORMS} or a number, not {norm}.")
    if norm == "deposited_energy" or norm == "deposited_energy_per_length":
        energy = True
        if rescaled_beam_intensity is None:
            raise ValueError("When norm is 'deposited_energy', beam_intensity " \
                           + "or rescaled_beam_intensity must be provided.")
    if normalise_by_length is None:
        if norm == "raw" or norm == "deposited_energy":
            normalise_by_length = False
        else:
            normalise_by_length = True
    elif norm == 'raw' and normalise_by_length:
        normalise_by_length = False
        print("Warning: norm is 'raw', normalise_by_length set to False.")
    elif norm == 'deposited_energy' and normalise_by_length:
        normalise_by_length = False
        print("Warning: norm is 'deposited_energy', normalise_by_length set to False.")
    elif norm == 'deposited_energy_per_length' and not normalise_by_length:
        normalise_by_length = True
        print("Warning: norm is 'deposited_energy_per_length', normalise_by_length set to True.")

    # Create arrays for plotting
    coll_s = lossmap['collimator']['s']
    coll_val = lossmap['collimator']['e'] if energy else lossmap['collimator']['n']
    if lossmap['interpolation']:
        aper_s = lossmap['aperture']['s_bins']
        aper_val = lossmap['aperture']['e_bins'] if energy else lossmap['aperture']['n_bins']
        aper_length = lossmap['aperture']['length_bins']
    else:
        aper_s = lossmap['aperture']['s']
        aper_val = lossmap['aperture']['e'] if energy else lossmap['aperture']['n']
        aper_length = lossmap['aperture']['length']
    if len(coll_s) == 0 and len(aper_s) == 0:
        raise ValueError("Empty loss map.")

    L = lossmap['machine_length']

    cold_s = np.array([], dtype=aper_s.dtype)
    cold_val = np.array([], dtype=aper_val.dtype)
    cold_length = 1
    warm_s = np.array([], dtype=aper_s.dtype)
    warm_val = np.array([], dtype=aper_val.dtype)
    warm_length = 1
    if cold_regions is not None:
        if warm_regions is not None:
            raise ValueError("Provide only one of cold_regions or warm_regions.")
        cold_regions = _unwrap_and_check(cold_regions, L)
        starts = np.sort(cold_regions[:, 0])
        ends   = np.sort(cold_regions[:, 1])
        num_starts = np.searchsorted(starts, aper_s, side='right') # For each s, count how many intervals start before or at s
        num_ends   = np.searchsorted(ends, aper_s, side='left')    # Count how many intervals end before s
        mask = num_starts > num_ends
        cold_s = aper_s[mask]
        cold_val = aper_val[mask]
        cold_length = aper_length if len(aper_length)==1 else aper_length[mask]
        warm_s = aper_s[~mask]
        warm_val = aper_val[~mask]
        warm_length = aper_length if len(aper_length)==1 else aper_length[~mask]
        aper_s = np.array([], dtype=aper_s.dtype)
        aper_val = np.array([], dtype=aper_val.dtype)
        aper_length = 1
    elif warm_regions is not None:
        warm_regions = _unwrap_and_check(warm_regions, L)
        starts = np.sort(warm_regions[:, 0])
        ends   = np.sort(warm_regions[:, 1])
        num_starts = np.searchsorted(starts, aper_s, side='right') # For each s, count how many intervals start before or at s
        num_ends   = np.searchsorted(ends, aper_s, side='left')    # Count how many intervals end before s
        mask = num_starts > num_ends
        warm_s = aper_s[mask]
        warm_val = aper_val[mask]
        warm_length = aper_length if len(aper_length)==1 else aper_length[mask]
        cold_s = aper_s[~mask]
        cold_val = aper_val[~mask]
        cold_length = aper_length if len(aper_length)==1 else aper_length[~mask]
        aper_s = np.array([], dtype=aper_s.dtype)
        aper_val = np.array([], dtype=aper_val.dtype)
        aper_length = 1

    if norm == "total":
        scale = coll_val.sum() + cold_val.sum() + warm_val.sum() + aper_val.sum()
    elif norm == "coll_max":
        scale = coll_val.max()
    elif norm == "max":
        scale = np.concatenate([coll_val, aper_val, cold_val, warm_val]).max()
    elif norm == "deposited_energy" or norm == "deposited_energy_per_length":
        scale = 1 / (rescaled_beam_intensity * 1.60218e-19)
    elif norm == "none" or norm == 'raw':
        scale = 1
    else:  # numeric
        scale = norm
    cold_val = cold_val / scale
    warm_val = warm_val / scale
    aper_val = aper_val / scale
    coll_val = coll_val / scale
    if normalise_by_length:
        cold_val = cold_val / cold_length
        warm_val = warm_val / warm_length
        aper_val = aper_val / aper_length
        coll_val = coll_val / lossmap['collimator']['length']

    if xlim is None:
        xlim = [-0.01*L, 1.01*L]
    if xlim[1] < xlim[0]:
        xlim[0] -= L
        cold_s = np.concatenate([cold_s - L, cold_s])
        warm_s = np.concatenate([warm_s - L, warm_s])
        aper_s = np.concatenate([aper_s - L, aper_s])
        coll_s = np.concatenate([coll_s - L, coll_s])
        cold_val = np.concatenate([cold_val, cold_val])
        warm_val = np.concatenate([warm_val, warm_val])
        aper_val = np.concatenate([aper_val, aper_val])
        coll_val = np.concatenate([coll_val, coll_val])
    if ylim is None:
        minimum = np.concatenate([coll_val, aper_val, cold_val, warm_val]).min()
        maximum = np.concatenate([coll_val, aper_val, cold_val, warm_val]).max()
        # ylim = [1.e-7, 1.e1]
        ylim = [minimum * 0.9 if minimum>0 else 1.e-9, maximum * 1.5]

    if grid:
        ax.grid(axis="y", which="major", zorder=0)

    bar_common_kwargs = dict(width = 0.8, lw = 1, bottom = 1.e-9)
    if len(coll_s) > 0:
        ax.bar(coll_s, coll_val, color="k", edgecolor="k", label="Collimator", zorder=9, **bar_common_kwargs)
    if len(aper_s) > 0:
        ax.bar(aper_s, aper_val, color="tab:orange", edgecolor="tab:orange", label="Aperture", zorder=10, **bar_common_kwargs)
    if len(warm_s) > 0:
        ax.bar(warm_s, warm_val, color="tab:red", edgecolor="tab:red", label="Warm",  zorder=11, **bar_common_kwargs)
    if len(cold_s) > 0:
        ax.bar(cold_s, cold_val, color="blue", edgecolor="blue", label="Cold",  zorder=12, **bar_common_kwargs)

    ax.set_yscale("log")
    ax.set_ylim(ylim)
    ax.set_xlim(xlim)

    if xticks is not None:
        ax.set_xticks(xticks)
    if yticks is None:
        yticks = [10**i for i in range(int(np.log10(ylim[0])), 1 + int(np.log10(ylim[1])))]
    ax.set_yticks(yticks)
    if x_label is None:
        x_label = "s [m]"
    ax.set_xlabel(x_label)
    if y_label is None:
        y_label = _label_from_norm(norm, energy)
    ax.set_ylabel(y_label)

    if legend:
        ax.legend(loc='upper right', fancybox=True, framealpha=0.8, fontsize="small")



def _valid_single_zoom(zz, allow_str=False):
    if isinstance(zz, str):
        return allow_str
    return hasattr(zz, "__iter__")  and len(zz) == 2 and not hasattr(zz[0], "__iter__") \
           and not hasattr(zz[1], "__iter__")

def _resolve_zoom(zoom, allow_str=False):
    if _valid_single_zoom(zoom, allow_str=allow_str):
        zoom = [zoom]
    if zoom is None:
        zoom = []
    if not hasattr(zoom, "__iter__") or isinstance(zoom, str) \
    or np.any([not _valid_single_zoom(zz, allow_str=allow_str) for zz in zoom]):
        raise ValueError("Zoom must be a list of 2 elements [s_min, s_max], or a list of such lists.")
    return zoom


def _label_from_norm(norm, energy):
    if norm == "raw":
        if energy:
            return "Deposited energy [eV]"
        else:
            return "Particles absorbed [-]"
    elif norm == "deposited_energy":
        return "Deposited energy [J]"
    elif norm == "deposited_energy_per_length":
        return "Deposited energy [J/m]"
    elif norm == "none":
        if energy:
            return "Deposited energy [eV/m]"
        else:
            return "Particles absorbed [1/m]"
    elif norm == "total":
        if energy:
            return "Norm. energy [1/m]"
        else:
            return "Norm. inefficiency [1/m]"
    elif norm == "max":
        if energy:
            return "Energy (norm. by max) [1/m]"
        else:
            return "Inefficiency (norm. by max) [1/m]"
    elif norm == "coll_max":
        if energy:
            return "Energy (norm. by coll max) [1/m]"
        else:
            return "Inefficiency (norm. by coll max) [1/m]"
    else:
        if energy:
            return f"Deposited energy [eV/m]"
        else:
            return f"Inefficiency [1/m]"


def _unwrap_and_check(pairs, L):
    # 1) unwrap to linear intervals in [0, L]
    linear = []
    for a, b in pairs:
        if a <= b:
            linear.append([a, b])
        else:
            linear.append([a, L])
            linear.append([0, b])

    if not linear:
        return []

    # 2) sort + merge on [0, L]
    linear.sort(key=lambda x: (x[0], x[1]))
    merged = [linear[0]]

    for a, b in linear[1:]:
        last = merged[-1]
        overlaps = a <= last[1]
        if overlaps:
            last[1] = max(last[1], b)
        else:
            merged.append([a, b])

    return np.array(merged)


# def _plot_arrow_beam_direction(lossmap, axes, beam_direction, maximum):
#     if type(beam_direction) is not dict:
#         beam_direction = {}
#     # arrow style
#     if 'arrow_opts' in beam_direction.keys():
#         arrow_opts = beam_direction['arrow_opts']
#     else:
#         arrow_opts = dict(color           = 'black',
#                           arrowstyle      = 'simple, head_width=.75, head_length=.75',
#                           connectionstyle = 'arc3,rad=0')
#     annotate_kwargs = dict(xycoords   = 'data',
#                            textcoords = 'data',
#                            arrowprops = arrow_opts,
#                            size       = 15)
#     arrow_hpos = 3000/27000*lossmap['machine_length']
#     arrow_vpos = 2.e-1 * maximum

#     # draw arrow
#     if lossmap.beam.name == "BEAM1":
#         text_alignment = 'left'
#         annotate_kwargs['xy']     = (arrow_hpos+1500, arrow_vpos)
#         annotate_kwargs['xytext'] = (arrow_hpos,      arrow_vpos)
#         xtext = arrow_hpos

#     elif lossmap.beam.name == "BEAM2":
#         annotate_kwargs['xy']     = (arrow_hpos     , arrow_vpos)
#         annotate_kwargs['xytext'] = (arrow_hpos+1500, arrow_vpos)
#         text_alignment = 'right'
#         xtext = arrow_hpos + 1500

#     axes[0].text(s="Beam", x=xtext, y=1.5*arrow_vpos, horizontalalignment=text_alignment, fontsize=17)
#     axes[0].annotate("", **annotate_kwargs)



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from contextlib import contextmanager

import xobjects as xo
import xtrack as xt

from .base import BaseCollimator, BaseCrystal
from ..general import _pkg_root
from ..scattering_routines.fluka import track_pre, track_core, track_post, FlukaEngine, \
                                        FlukaPrototype, create_generic_assembly
from ..materials import _resolve_material
from ..constants import HIT_ON_FLUKA_COLL


class FlukaCollimator(BaseCollimator):
    _xofields = { **BaseCollimator._xofields,
        'fluka_id':              xo.Int16,    # Do not change! Should be 16 bit because of FlukaIO type
        'length_front':          xo.Float64,
        'length_back':           xo.Float64,
        '_tracking':             xo.Int8,
        '_acc_ionisation_loss':  xo.Float64,  # TODO: this is not very robust, for when a track is done with new particles etc
    }

    isthick = True
    allow_track = True
    iscollective = True
    behaves_like_drift = True
    allow_rot_and_shift = False
    skip_in_loss_location_refinement = True

    _depends_on = [BaseCollimator, FlukaEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','fluka_collimator.h')
    ]

    _noexpr_fields         = {*BaseCollimator._noexpr_fields, 'material', 'assembly'}
    _skip_in_to_dict       = [*BaseCollimator._skip_in_to_dict, '_tracking', '_acc_ionisation_loss']
    _store_in_to_dict      = [*BaseCollimator._store_in_to_dict, 'material', 'assembly', 'height', 'width', 'side']
    _internal_record_class = BaseCollimator._internal_record_class

    _allowed_fields_when_frozen = ['_tracking', '_acc_ionisation_loss']

    def __new__(cls, *args, **kwargs):
        with cls._in_constructor():
            self = super().__new__(cls, *args, **kwargs)
        return self

    def __init__(self, **kwargs):
        import xcoll as xc
        if xc.fluka.engine.is_running():
            raise ValueError('Cannot create FlukaCollimator while engine is running.')
        with self.__class__._in_constructor(self):
            to_assign = {}
            if '_xobject' not in kwargs:
                kwargs.setdefault('_tracking', True)
                kwargs.setdefault('_acc_ionisation_loss', -1.)
                to_assign['name'] = xc.fluka.engine._get_new_element_name()
                assembly = kwargs.pop('assembly', None)
                material = _resolve_material(kwargs.pop('material', None), ref='fluka', allow_none=True)
                side = kwargs.pop('side', None)
                width = kwargs.pop('width', None)
                height = kwargs.pop('height', None)
                if assembly is not None:
                    # Use the provided assembly, check consistency later
                    generic = False
                    to_assign['assembly'] = assembly
                else:
                    # Create a generic assembly
                    generic = True
                    if material is None:
                        raise ValueError('Need to provide material!')
                    length = kwargs.get('length', None)
                    if length is None:
                        raise ValueError('Need to provide length!')
            self._assembly = None  # Temporary, to avoid issues in setters
            super().__init__(**kwargs)
            for key, val in to_assign.items():
                setattr(self, key, val)
            if generic:
                side = self._get_side_from_input(side)
                self.assembly = create_generic_assembly(material=material,
                                    side=side, length=self.length, width=width,
                                    height=height)
            else:
                # Check consistency
                if self.assembly.material is not None and material is not None \
                and self.assembly.material != material:
                    print("Material in assembly and provided `material` "
                        + "argument do not agree! Ignored the latter.")
                if self.assembly.side is not None and side is not None \
                and self.assembly.side != side:
                    print("Side in assembly and provided `side` "
                        + "argument do not agree! Ignored the latter.")
                if self.assembly.width is not None and width is not None \
                and self.assembly.width != width:
                    print("Width in assembly and provided `width` "
                        + "argument do not agree! Ignored the latter.")
                if self.assembly.height is not None and height is not None \
                and self.assembly.height != height:
                    print("Height in assembly and provided `height` "
                        + "argument do not agree! Ignored the latter.")
            if not hasattr(self, '_equivalent_drift'):
                self._equivalent_drift = xt.Drift(length=self.length)
                self._equivalent_drift.model = 'exact'

    @property
    def material(self):
        if self.assembly is not None:
            return self.assembly.material

    @material.setter
    def material(self, material):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change material of non-generic assembly!')
            material = _resolve_material(material, ref='fluka', allow_none=False)
            if self.material != material:
                self.assembly = create_generic_assembly(material=material,
                                side=self.side, length=self.length, width=self.width,
                                height=self.height)

    @property
    def height(self):
        if self.assembly is not None:
            return self.assembly.height

    @height.setter
    def height(self, height):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change height of non-generic assembly!')
            self.assembly = create_generic_assembly(material=self.material,
                            side=self.side, length=self.length, width=self.width,
                            height=height)

    @property
    def width(self):
        if self.assembly is not None:
            return self.assembly.width

    @width.setter
    def width(self, width):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change width of non-generic assembly!')
            self.assembly = create_generic_assembly(material=self.material,
                            side=self.side, length=self.length, width=width,
                            height=self.height)
    @property
    def side(self):
        if self.assembly is not None:
            return self.assembly.side

    @side.setter
    def side(self, side):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change side of non-generic assembly!')
            side = self._get_side_from_input(side)
            self.assembly = create_generic_assembly(material=self.material,
                            side=side, length=self.length, width=self.width,
                            height=self.height)

    @property
    def assembly(self):
        if self._assembly is not None:
            if self._assembly.is_defunct() or self._assembly._is_null:
                self._assembly = None
        return self._assembly

    @assembly.setter
    def assembly(self, val):
        import xcoll as xc
        if isinstance(val, str):
            if val in xc.fluka.assemblies:
                val = xc.fluka.assemblies[val]
            elif val in xc.fluka.prototypes:
                val = xc.fluka.prototypes[val]
            else:
                raise ValueError(f"Unknown assembly/prototype '{val}'.")
        elif not isinstance(val, FlukaPrototype) and val is not None:
            raise ValueError(f'Invalid assembly/prototype {val}!')
        if val.is_broken:
            print(f'Warning: assembly/prototype {val.name} is broken!')
        if val.is_defunct():
            raise ValueError(f'Cannot assign defunct assembly/prototype {val.name}!')
        val.assert_exists()
        self._assembly = val
        if self.assembly:
            if self.assembly.length is not None:
                self.length_front = (self.assembly.length - self.length) / 2
                self.length_back = self.assembly.length - self.length - self.length_front
            if self.assembly.side is not None:
                self._get_side_from_input(self.assembly.side)

    def enable_scattering(self):
        import xcoll as xc
        xc.fluka.environment.assert_environment_ready()
        if not xc.fluka.engine.is_running():
            raise RuntimeError("FLUKA engine is not running.")
        super().enable_scattering()

    def track(self, part):
        if track_pre(self, part):
            if self.material != "vacuum" and False:
                super().track(part)
            else:
                part.state[part.state == 1] = HIT_ON_FLUKA_COLL
            track_core(self, part)
            if self.material != "vacuum":
                track_post(self, part)
        else:
            self._drift(part)

    def _drift(self, particles, length=None):
        if length is None:
            length = self.length
        if length != self.length:
            old_length = self._equivalent_drift.length
            self._equivalent_drift.length = length
        self._equivalent_drift.track(particles)
        if length != self.length:
            self._equivalent_drift.length = old_length

    def __setattr__(self, name, value):
        import xcoll as xc
        if name not in self._allowed_fields_when_frozen \
        and xc.fluka.engine.is_running() is True:
            raise ValueError('Engine is running; FlukaCollimator is frozen.')
        super().__setattr__(name, value)

    @classmethod
    @contextmanager
    def _in_constructor(cls, self=None):
        original_setattr = cls.__setattr__
        if self is not None:
            self._being_constructed_ = True
        def new_setattr(self, *args, **kwargs):
            return super().__setattr__( *args, **kwargs)
        cls.__setattr__ = new_setattr
        try:
            yield
        finally:
            cls.__setattr__ = original_setattr
            if self is not None:
                self._being_constructed_ = False

    def _being_constructed(self):
        if hasattr(self, '_being_constructed_'):
            return self._being_constructed_
        else:
            return False

    # ===================================================
    # ===   Hacks to use parent setters and getters   ===
    # ===================================================

    def _get_side_from_input(self, side):
        # Set / get using _side field (even though it is not used directly)
        if side is not None:
            BaseCollimator.side.fset(self, side)
            side = BaseCollimator.side.fget(self)
            return side



class FlukaCrystal(BaseCrystal):
    _xofields = { **BaseCrystal._xofields,
        'fluka_id':              xo.Int16,    # Do not change! Should be 16 bit because of FlukaIO type
        'length_front':          xo.Float64,
        'length_back':           xo.Float64,
        '_tracking':             xo.Int8,
        '_acc_ionisation_loss':  xo.Float64,  # TODO: this is not very robust, for when a track is done with new particles etc
    }

    isthick = True
    allow_track = True
    iscollective = True
    behaves_like_drift = True
    allow_rot_and_shift = False
    skip_in_loss_location_refinement = True

    _noexpr_fields         = {*BaseCrystal._noexpr_fields, 'material', 'assembly'}
    _skip_in_to_dict       = [*BaseCrystal._skip_in_to_dict, '_tracking', '_acc_ionisation_loss']
    _store_in_to_dict      = [*BaseCrystal._store_in_to_dict, 'material', 'assembly', 'height', 'width', 'side']
    _internal_record_class = BaseCrystal._internal_record_class

    _depends_on = [BaseCrystal, FlukaEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','fluka_crystal.h')
    ]

    _allowed_fields_when_frozen = ['_tracking', '_acc_ionisation_loss']

    def __new__(cls, *args, **kwargs):
        with cls._in_constructor():
            self = super().__new__(cls, *args, **kwargs)
        return self

    def __init__(self, **kwargs):
        import xcoll as xc
        if xc.fluka.engine.is_running():
            raise ValueError('Cannot create FlukaCrystal while engine is running.')
        with self.__class__._in_constructor(self):
            to_assign = {}
            generic = False
            if '_xobject' not in kwargs:
                kwargs.setdefault('_tracking', True)
                kwargs.setdefault('_acc_ionisation_loss', -1.)
                to_assign['name'] = xc.fluka.engine._get_new_element_name()
                assembly = kwargs.pop('assembly', None)
                if assembly:
                    raise NotImplementedError('FlukaCrystalAssemblies not yet implemented!')
                # kwargs for generic assembly creation
                material = _resolve_material(kwargs.pop('material', None), ref='fluka', allow_none=True)
                side = kwargs.pop('side', None)
                width = kwargs.pop('width', None)
                height = kwargs.pop('height', None)
                bending_radius = kwargs.pop('bending_radius', None)
                bending_angle = kwargs.pop('bending_angle', None)
                if assembly is not None:
                    # Use the provided assembly, check consistency later
                    generic = False
                    to_assign['assembly'] = assembly
                else:
                    # Create a generic assembly
                    generic = True
                    if material is None:
                        raise ValueError('Need to provide material!')
                    length = kwargs.get('length', None)
                    if length is None:
                        raise ValueError('Need to provide length!')
                    if side is None:
                        raise ValueError('Need to provide side!')
                    if bending_radius is None and bending_angle is None:
                        raise ValueError('Need to provide bending radius or angle!')
                    elif bending_radius is not None and bending_angle is not None:
                        raise ValueError('Cannot provide both bending radius and angle!')
                # Some defaults to keep BaseCrystal happy
                kwargs.setdefault('_bending_radius', 1.)
                length = kwargs.get('length', kwargs['length'] or 1 )
                kwargs.setdefault('_bending_angle', np.arcsin(length))
            self._assembly = None  # Temporary, to avoid issues in setters
            super().__init__(**kwargs)
            for key, val in to_assign.items():
                setattr(self, key, val)
            if generic:
                if bending_radius is None:
                    bending_radius = self._get_bending_radius_from_angle(bending_angle)
                else:
                    self._get_bending_angle_from_radius(bending_radius)  # To set internal fields correctly
                side = self._get_side_from_input(side)
                self.assembly = create_generic_assembly(is_crystal=True, material=material,
                                side=side, length=self.length, width=width, height=height,
                                bending_radius=bending_radius)
            else:
                # Check consistency
                if self.assembly.material is not None and material is not None \
                and self.assembly.material != material:
                    print("Material in assembly and provided `material` "
                        + "argument do not agree! Ignored the latter.")
                if self.assembly.side is not None and side is not None \
                and self.assembly.side != side:
                    print("Side in assembly and provided `side` "
                        + "argument do not agree! Ignored the latter.")
                if self.assembly.width is not None and width is not None \
                and self.assembly.width != width:
                    print("Width in assembly and provided `width` "
                        + "argument do not agree! Ignored the latter.")
                if self.assembly.height is not None and height is not None \
                and self.assembly.height != height:
                    print("Height in assembly and provided `height` "
                        + "argument do not agree! Ignored the latter.")
                if self.assembly.bending_radius is not None and bending_radius is not None \
                and self.assembly.bending_radius != bending_radius:
                    print("Bending_radius in assembly and provided `bending_radius` "
                        + "argument do not agree! Ignored the latter.")
                # TODO: check bending_angle consistency too
            if not hasattr(self, '_equivalent_drift'):
                self._equivalent_drift = xt.Drift(length=self.length)
                self._equivalent_drift.model = 'exact'

    @property
    def material(self):
        return FlukaCollimator.material.fget(self)

    @material.setter
    def material(self, material):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change material of non-generic assembly!')
            material = _resolve_material(material, ref='fluka', allow_none=False)
            if self.material != material:
                self.assembly = create_generic_assembly(is_crystal=True, material=material,
                                side=self.side, length=self.length, width=self.width,
                                height=self.height, bending_radius=self.bending_radius)

    @property
    def height(self):
        return FlukaCollimator.height.fget(self)

    @height.setter
    def height(self, height):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change height of non-generic assembly!')
            self.assembly = create_generic_assembly(is_crystal=True, material=self.material,
                            side=self.side, length=self.length, width=self.width,
                            height=height, bending_radius=self.bending_radius)

    @property
    def width(self):
        return FlukaCollimator.width.fget(self)

    @width.setter
    def width(self, width):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change width of non-generic assembly!')
            self.assembly = create_generic_assembly(is_crystal=True, material=self.material,
                            side=self.side, length=self.length, width=width,
                            height=self.height, bending_radius=self.bending_radius)

    @property
    def side(self):
        if self.assembly is not None:
            return FlukaCollimator.side.fget(self)

    @side.setter
    def side(self, side):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change side of non-generic assembly!')
            side = self._get_side_from_input(side)
            self.assembly = create_generic_assembly(is_crystal=True, material=self.material,
                            side=side, length=self.length, width=self.width,
                            height=self.height, bending_radius=self.bending_radius)
    @property
    def bending_radius(self):
        if self.assembly is not None:
            return self.assembly.bending_radius

    @bending_radius.setter
    def bending_radius(self, bending_radius):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change bending radius of non-generic assembly!')
            self._get_bending_angle_from_radius(bending_radius) # To set internal fields correctly
            self.assembly = create_generic_assembly(is_crystal=True, material=self.material,
                            side=self.side, length=self.length, width=self.width,
                            height=self.height, bending_radius=bending_radius)

    @property
    def bending_angle(self):
        return self._get_bending_angle_from_radius(self.bending_radius)

    @bending_angle.setter
    def bending_angle(self, bending_angle):
        if not self._being_constructed():
            if self.assembly.fedb_series != 'generic':
                raise ValueError('Cannot change bending angle of non-generic assembly!')
            self.bending_radius = self._get_bending_radius_from_angle(bending_angle)

    @property
    def assembly(self):
        return self._assembly

    @assembly.setter
    def assembly(self, assembly):
        FlukaCollimator.assembly.fset(self, assembly)
        if self.assembly and not self.assembly.is_crystal:
            raise ValueError('Assigned assembly is not a crystal assembly!')

    def enable_scattering(self):
        import xcoll as xc
        xc.fluka.environment.assert_environment_ready()
        if not xc.fluka.engine.is_running():
            raise RuntimeError("FLUKA engine is not running.")
        super().enable_scattering()

    def track(self, part):
        if track_pre(self, part):
            # super().track(part)
            part.state[part.state == 1] = HIT_ON_FLUKA_COLL
            track_core(self, part)
            track_post(self, part)
        else:
            self._drift(part)

    def _drift(self, particles, length=None):
        if length is None:
            length = self.length
        if length != self.length:
            old_length = self._equivalent_drift.length
            self._equivalent_drift.length = length
        self._equivalent_drift.track(particles)
        if length != self.length:
            self._equivalent_drift.length = old_length

    def __setattr__(self, name, value):
        import xcoll as xc
        if name not in self._allowed_fields_when_frozen \
        and xc.fluka.engine.is_running() is True:
            raise ValueError('Engine is running; FlukaCrystal is frozen.')
        super().__setattr__(name, value)

    _in_constructor = FlukaCollimator._in_constructor
    _being_constructed = FlukaCollimator._being_constructed


    # ===================================================
    # ===   Hacks to use parent setters and getters   ===
    # ===================================================

    def _get_side_from_input(self, side):
        # Set / get using _side field (even though it is not used directly)
        if side is not None:
            BaseCrystal.side.fset(self, side)
            side = BaseCrystal.side.fget(self)
            return side

    def _get_bending_radius_from_angle(self, bending_angle):
        # Set / get using _bending_radius and _bending_angle fields (even though they are not used directly)
        if self.assembly:
            old_length = self.length
            self.length = self.assembly.length
        BaseCrystal.bending_angle.fset(self, bending_angle)
        bending_radius = BaseCrystal.bending_radius.fget(self)
        if self.assembly:
            self.length = old_length
        return bending_radius

    def _get_bending_angle_from_radius(self, bending_radius):
        # Set / get using _bending_radius and _bending_angle fields (even though they are not used directly)
        if self.assembly:
            old_length = self.length
            self.length = self.assembly.length
        BaseCrystal.bending_radius.fset(self, bending_radius)
        bending_angle = BaseCrystal.bending_angle.fget(self)
        if self.assembly:
            self.length = old_length
        return bending_angle



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #


class XcollAccessor:
    """A uniform way to access elements from an underlying database.

    The class provides dictionary-like access to elements, and element
    attributes can be retrieved and set directly on the accessor,
    providing a dict(element -> attribute) interface.

    The class expects an attribute '_db' (which can be anything that has
    a 'get' method) that holds the elements. Furthermore, the class needs
    'names', which can be a (potentially dynamic) property or set as an
    attribute in the init, to specify which elements can be addressed by
    the accessor. If not specified, all elements in the db are considered
    (in which case names is equal to self._db.keys()).

    The attributes inside the elements are expected to be accessed as an
    item (if the element is a dict), as an attribute (any other object),
    or via a get method (if the previous two lookups failed). Similarly,
    when setting an attribute, the accessor will first try to set it
    as an item, then as an attribute, and finally via a set method if
    available (otherwise it fails).

    Additionally, each element can have an attribute 'family' that gives
    the name of the family it belongs to, and 'non_family_attributes'
    which lists its attributes that will not be updated when the family
    attribute is set. This will auto-generate a families dict, and
    sub-accessors when accessing by family name.

    Accessing elements:
        Elements of the database can be accessed using the standard
        subscription syntax (e.g. accessor['name'] which returns the
        element itself). If `name` is a family name instead, it will
        return a sub-accessor for that family. The subscription syntax
        cannot be used to generated new elements (i.e.
            accessor['new_element'] = ...
        does not work), but it can be used to set attributes for an
        element with a settings dict:
            acc['el1'] = {'a': -10, 'd': -10, 'f': -999}
        where non-specified attributes will be left unchanged and missing
        ones are generated. When using this approach on a family name,
        all elements in the family will be updated, except for attributes
        that are listed in the non_family_attributes of an element.

    Accessing attributes:
        When accessing an attribute on the accessor, a dict of element
        names and the value of the respective attribute is returned.
        Elements that do not have the requested attribute are omitted. If
        all elements have the same value for the requested attribute,
        this value is returned directly instead of a dict. If no elements
        have the requested attribute, an AttributeError is raised.

    Accessor API:
        - accessor.names: list of element names that are represented by
            the accessor (manually set or dynamic property).
        - accessor.families: dict of family name -> list of element names
            in that family.
        - accessor.family_names: list of family names.
        - accessor.show(): print the content of the accessor.
        - iteration over the accessor iterates over the elements in the
            accessor (unlike a dict which iterates over the keys). The
            elements are returned in the same order as they appear in
            accessor.names.
        - accessor.keys(), accessor.values(), accessor.items() work as
            for a dict.


    Setting an attribute on the accessor will set it for all elements, unless the attribute is listed in an element's non_family_attributes, in which case it will be ignored for that element. Setting an item on the accessor will set attributes for that element, or for all elements in the family if a family name is used as key.
    """

    # These are only used for printing and error messages:
    _typename = 'element'   # What are we accessing? (e.g. elements, collimators, ..)
    _dbtype   = 'line'      # What is the underlying db? (e.g. line, colldb, ..)
    _eltype   = None        # What type of elements are we accessing? (e.g. BeamElement, settings, ..)

    def __init__(self, db, names=None, **kwargs_to_set):
        super().__setattr__('_db', db)
        kwargs_to_set.setdefault('_is_family_sub_accessor', False)
        if names:
            super().__setattr__('names', names)
        elif not 'names' in self.__class__.__dict__:
            super().__setattr__('names', list(db.keys()))
        for key, value in kwargs_to_set.items():
            super().__setattr__(key, value)
        self._check_family_consistency()

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))} (use .show() " \
             + f"to see the content)>"

    def __str__(self):
        if len(self.names) == 0:
            return ''
        res = []
        if len(self.families) > 0:
            res.append('Families:')
            name_len = max(len(name) for name in self.family_names) + 1
            name_len = max(name_len, 10)
            for family, names in self.families.items():
                ff = f'{family}:'
                res.append(f"    {ff:<{name_len}}  {', '.join(names)}")
            fam_names = {vvv for vv in self.families.values() for vvv in vv}
            nofam_names = set(self.names) - fam_names
            if nofam_names:
                head = f"{'no family:':<{name_len}}"
                res.append(f"    {head}  {', '.join(nofam_names)}")
            res.append('')
        res.append(f'{self._typename.capitalize()}s:')
        name_len = max(len(name) for name in self.names)
        for name in self.names:
            if self._eltype:
                cls_name = self._eltype
            else:
                cls_name = self._element_dict[name].__class__.__name__
            res.append(f"    {name:<{name_len}}  ({cls_name})  {self[name]}")
        return "\n".join(res)

    def show(self):
        """Print the content of the accessor."""
        print(self)

    @property
    def families(self):
        families = {}
        try:
            prop_families = self.family
        except AttributeError:
            return families
        else:
            for name in self.names:
                if name in prop_families:
                    if prop_families[name] not in families:
                        families[prop_families[name]] = []
                    families[prop_families[name]].append(name)
            return families

    @property
    def family_names(self):
        return list(self.families.keys())

    @property
    def _element_dict(self):
        # Helper attribute to get correct class instances for keys,
        # values, and items methods
        return {name: self._db.get(name) for name in self.names}

    def keys(self):
        return self._element_dict.keys()

    def values(self):
        return self._element_dict.values()

    def items(self):
        return self._element_dict.items()

    def __iter__(self):
        super().__setattr__('_iter_names', iter(self.names))
        return self

    def __next__(self):
        try:
            name = next(self._iter_names)
        except StopIteration:
            raise StopIteration
        else:
            # Use _get_db_element instead of _db.get to protect against
            # potential changes in between different calls to __next__.
            return self._get_db_element(name)

    def __len__(self):
        return len(self.names)

    def __contains__(self, key):
        return key in self.names

    # @property   # This only works if the underlying db is a dict...
    # def properties(self):
    #     return {attr for ddd in self.values() for attr in ddd.keys()}

    def _get_db_element(self, name):
        if name in self.names:
            return self._db.get(name)
        else:
            raise KeyError(f"{self._typename.capitalize()} `{name}` not "
                           f"found in {self._dbtype}!")

    def _get_element_attr(self, name, attr, default=None,
                          allow_missing=False):
        if default is not None:
            # If a default is provided, we allow missing attributes
            allow_missing = True
        el = self._get_db_element(name)
        if isinstance(el, dict):
            if attr in el:
                return el[attr]
            elif not allow_missing:
                raise AttributeError(f"Attribute `{attr}` not found in "
                                     f"{self._typename} `{name}`!")
            else:
                return default
        elif hasattr(el, attr):
            return getattr(el, attr)
        else:
            try:
                return el.get(attr, default)
            except:
                if not allow_missing:
                    raise AttributeError(f"Attribute `{attr}` not found in "
                                         f"{self._typename} `{name}`!")
                return default

    def _set_element_attr(self, name, attr, value, allow_missing=True,
                          add_new=False):
        el = self._get_db_element(name)
        if not allow_missing:
            if add_new:
                raise ValueError("Cannot use `add_new=True` and "
                                 "`allow_missing=False` at the same time!")
            # Check that the attribute exists, raise error if not
            self._get_element_attr(name, attr, allow_missing=False)
        if isinstance(el, dict):
            if attr in el or add_new:
                el[attr] = value
        elif hasattr(el, attr) or add_new:
            try:
                setattr(el, attr, value)
            except:
                try:
                    el.set(attr, value)
                except:
                    raise AttributeError(f"Cannot set attribute `{attr}` in "
                                         f"{self._typename} `{name}`!")

    def _get_element_non_family_attributes(self, name):
        return self._get_element_attr(name, 'non_family_attributes',
                                      default=[])

    def _element_has_attr(self, name, attr):
        prop = self._get_element_attr(name, attr,
                                      default='__xc_acc_not_found__')
        return prop != '__xc_acc_not_found__'

    def __getattr__(self, attr):
        properties = {}
        for name in self.names:
            if self._element_has_attr(name, attr):
                properties[name] = self._get_element_attr(name, attr)
        if len(properties) == 0:
            raise AttributeError(f"Attribute `{attr}` not found in "
                                 f"{self._dbtype}!")
        # If all values are the same, return a single value
        if len(properties) == len(self.names) \
        and len({tuple(ii) if isinstance(ii, list) else ii
                for ii in properties.values()}) == 1:
            return next(iter(properties.values()))
        # If this is a family accessor, all members should have it
        if self._is_family_sub_accessor:
            missing = set(self.names) - set(properties.keys())
            for name in missing:
                # It is fine if the attribute is declared non-family
                # in this element
                if attr not in self._get_element_non_family_attributes(name):
                    raise AttributeError(f"Attribute `{attr}` is not a family "
                        + f"attribute, as {self._typename} `{name}` does not "
                        + f"have it! If this is intentional, add the attribute"
                        + f" to `{name}`s non_family_attributes list!")
        return properties

    def __setattr__(self, attr, value):
        # First check that the attribute exists 
        getattr(self, attr)
        if isinstance(value, dict):
            if self._is_family_sub_accessor:
                raise ValueError("Cannot set attribute with a dict on a "
                                 "family!")
            for name, val in value.items():
                self._set_element_attr(name, attr, val, allow_missing=False)
        else:
            # If the value is not a dict, it is either a family attribute...
            if self._is_family_sub_accessor:
                for name in self.names:
                    if attr in self._get_element_non_family_attributes(name):
                        # This attribute is non-family, so ignore it when
                        # setting the family
                        continue
                    self._set_element_attr(name, attr, value)
            # ... or a single value to set for all elements
            else:
                for name in self.names:
                    self._set_element_attr(name, attr, value)

    def __getitem__(self, name):
        # We can getitem by name or family, so we overwrite the super method
        if name in self.families:
            return XcollAccessor(db=self._db, names=self.families[name],
                        _typename=self._typename, _dbtype=self._dbtype,
                        _is_family_sub_accessor=True)
        else:
            return self._get_db_element(name)

    def __setitem__(self, name, value):
        # Set attributes of a single element. This is the only way to add new
        # attributes to an element.
        # TODO: allow to define a family this way
        if name in self.families:
            # Set attributes for all family members
            if not isinstance(value, dict):
                raise ValueError(f"Can only set family `{name}` to a "
                               + f"settings dict!")
            for el_name in self.families[name]:
                for attr, val in value.items():
                    if attr in self._get_element_non_family_attributes(el_name):
                        # This attribute is non-family, so ignore it when setting the family
                        continue
                    self._set_element_attr(el_name, attr, val, add_new=True)
        else:
            if not isinstance(value, dict):
                raise ValueError(f"Can only set {self._typename} `{name}` to a "
                               + f"settings dict!")
            # Set attributes for this element
            for attr, val in value.items():
                self._set_element_attr(name, attr, val, add_new=True)

    def _check_family_consistency(self):
        for name in self.names:
            fam = self._get_element_attr(name, 'family', allow_missing=True)
            if fam is not None and not isinstance(fam, str):
                raise ValueError(f"Inconsistent db in XcollAccessor: "
                                 f"{self._typename.capitalize()} `{name}` has "
                                 f"a `family` attribute that is not a string!")
            nf = self._get_element_non_family_attributes(name)
            if nf:
                if fam is None:
                    raise AttributeError(f"Inconsistent db in XcollAccessor: "
                        f"{self._typename.capitalize()} `{name}` has "
                        f"`non_family_attributes` but no `family` attribute!")
                if isinstance(nf, str):
                    self._set_element_attr(name, 'non_family_attributes', [nf])
                elif not hasattr(nf, '__iter__'):
                    raise ValueError(f"Inconsistent db in XcollAccessor: "
                                     f"{self._typename.capitalize()} `{name}` "
                                     f"has `non_family_attributes` but it is "
                                     f"not iterable!")
                for attr in nf:
                    if not self._element_has_attr(name, attr):
                        raise AttributeError(f"Inconsistent db in "
                            f"XcollAccessor: {self._typename.capitalize()} "
                            f"`{name}` has attribute `{attr}` in "
                            f"`non_family_attributes` but it is not present!")
                    if attr in ['family', 'non_family_attributes']:
                        raise ValueError(f"Inconsistent db in XcollAccessor: "
                            f"{self._typename.capitalize()} `{name}` has "
                            f"`{attr}` in `non_family_attributes`!")


class XcollAccessorElementProxy:
    """A view on a single element of an XcollAccessor."""

    def __init__(self, accessor: XcollAccessor, name: str):
        self._accessor = accessor
        self._name = name
        self._element = accessor._get_db_element(name)

    def __str__(self):
        return self._element.__str__()

    def __repr__(self):
        return self._element.__repr__()

    def __getattr__(self, name):
        return getattr(self._target, name)

    def __getattr__(self, attr):
        return self._accessor._get_element_attr(self._name, attr)

    def __setattr__(self, attr, value):
        if attr in ['_accessor', '_name']:
            super().__setattr__(attr, value)
        else:
            self._accessor._set_element_attr(self._name, attr, value)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
import scipy.constants as sc


# Be careful! Changing those default values will change all materials that do not set them explicitly!
# Hence, when loading jsons saved before the change, the materials will change!

_default_excitation_energies = {
    1: 19.2, 2: 41.8, 3: 40, 4: 63.7, 5: 76, 6: 81, 7: 82, 8: 95, 9: 115, 10: 137, 11: 149, 12: 156, 13: 166,
    14: 173, 15: 173, 16: 180, 17: 174, 18: 188, 19: 190, 20: 191, 21: 216, 22: 233, 23: 245, 24: 257, 25: 272,
    26: 286, 27: 297, 28: 311, 29: 322, 30: 330, 31: 334, 32: 350, 33: 347, 34: 348, 35: 343, 36: 352, 37: 363,
    38: 366, 39: 379, 40: 393, 41: 417, 42: 424, 43: 428, 44: 441, 45: 449, 46: 470, 47: 470, 48: 469, 49: 488,
    50: 488, 51: 487, 52: 485, 53: 491, 54: 482, 55: 488, 56: 491, 57: 501, 58: 523, 59: 535, 60: 546, 61: 560,
    62: 574, 63: 580, 64: 591, 65: 614, 66: 628, 67: 650, 68: 658, 69: 674, 70: 684, 71: 694, 72: 705, 73: 718,
    74: 727, 75: 736, 76: 746, 77: 757, 78: 790, 79: 790, 80: 800, 81: 810, 82: 823, 83: 823, 84: 830, 85: 825,
    86: 794, 87: 827, 88: 826, 89: 841, 90: 847, 91: 878, 92: 890, 93: 902, 94: 921, 95: 934, 96: 939, 97: 952,
    98: 966, 100: 994, 116: 1213
}

def _approximate_radiation_length(Z, A, density):
    # Tsai's more accurate expression (PDG'24 Sec 34.4.2)
    match int(round(Z)):
        case 1:  # Hydrogen
            L_rad = 5.31
            L_rad_prime = 6.144
        case 2:  # Helium
            L_rad = 4.79
            L_rad_prime = 5.621
        case 3:  # Lithium
            L_rad = 4.74
            L_rad_prime = 5.805
        case 4:  # Beryllium
            L_rad = 4.71
            L_rad_prime = 5.924
        case _:
            L_rad = np.log(184.15 * Z ** (-1.0 / 3.0))
            L_rad_prime = np.log(1194.0 * Z ** (-2.0 / 3.0))
    denom = Z**2 * (L_rad - _f_coulomb(Z)) + Z * L_rad_prime
    X0_mass = 716.405 * A / denom  # in g/cm^2
    return X0_mass / density / 100  # in meters

def _f_coulomb(Z):
    # Coulomb correction f(Z) from PDG.
    a = sc.alpha * Z
    a2 = a*a
    return a2 * (
        1/(1+a2)
        + 0.20206
        - 0.0369*a2
        + 0.0083*a2*a2
        - 0.0020*a2*a2*a2
    )


# --- Mixture rules ------------------------------------------------------------

def _average_Z_over_A(components, mass_fractions):
    return sum([fr * el.Z / el.A for el, fr in zip(components,
                                                   mass_fractions)])

def _effective_Z2(components, molar_fractions):
    return np.sum([mf * el.Z**2 for el, mf in zip(components,
                                                  molar_fractions)])

def _combine_radiation_lengths(components, mass_fractions, density):
    # radiation_length in m
    X0_mass = [el.radiation_length * 100 * el.density for el in components]  # g/cm^2
    X0_mass_mix = _inverse_weighted_mean(X0_mass, mass_fractions)            # g/cm^2
    return X0_mass_mix / density / 100                                       # m

def _combine_excitation_energies(components, mass_fractions):
    fractions = np.array([fr * el.Z / el.A for el, fr in zip(components,
                                                             mass_fractions)])
    return _logarithmic_mean([el.excitation_energy for el in components],
                              fractions)



def _inverse_weighted_mean(vals, weights):
    return 1 / (np.array(weights) / np.array(vals)).sum()

def _logarithmic_mean(vals, weights):
    res = np.sum(np.array(weights) * np.log(np.array(vals))) / np.sum(weights)
    return np.exp(res)



import os
import sys
import argparse

def prRed(skk):  print("\033[91m {}\033[00m" .format(skk), end="")
def prCyan(skk): print("\033[96m {}\033[00m" .format(skk), end="")


def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-n", "--no_material_expansion", action="store_true")
    parser.add_argument("input", nargs='*', help="<FLUKA file> [expanded FLUKA file]")
    args = parser.parse_args()

    try:
        inputfile = args.input[0]
    except:
        parser.print_help()
        return 1
    assert inputfile.endswith(".inp"), "Input file should have .inp extension"

    if len(args.input) > 1:
        outputfile = args.input[1]
    else:
        outputfile = inputfile.replace(".inp","_exp.inp")

    lines = []
    file_expandable = False
    with open(inputfile) as ifile:
        for line in ifile:
            # file to be included: open and read it
            if line.startswith("#include"):
                data = line.strip().split()
                include_file = data[1]

                # do not expand materials.inp
                no_material_expansion = include_file.endswith("materials.inp") and args.no_material_expansion
                if os.path.isfile(include_file) and no_material_expansion:
                    lines.append(line)
                elif os.path.isfile(include_file):
                    prCyan("Including: ")
                    print(include_file)
                    with open(include_file, 'r') as fin:
                        lines.append(fin.read())
                        file_expandable = True
                else:
                    lines.append(line)
            # normal line
            else:
                lines.append(line)

    if not file_expandable:
        prRed(f"{inputfile} not expandable\n")
        return 0

    with open(outputfile, 'w') as ofile:
        for line in lines:
            if not line.strip(): continue
            if '\n' in line:
                ofile.write(f"{line}")
            else:
                ofile.write(f"{line}\n")
    return 0


if __name__ == "__main__":
    sys.exit(main())



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

# Tenporary file that defines xaux tools - to be used until xaux is in main Xsuite release cycle

import os
import base64
import shutil
import inspect
import functools
import numpy as np
from pathlib import PosixPath


class FsPath(PosixPath):
    def copy_to(self, other, **kwargs):
        if self.is_dir():
            shutil.copytree(self, other / self.name, dirs_exist_ok=True)
        else:
            shutil.copy(self, other)
    def move_to(self, other, **kwargs):
        shutil.move(self, other)
    def rmtree(self, *args, **kwargs):
        shutil.rmtree(self, *args, **kwargs)
    def __eq__(self, other):
        try:
            other = FsPath(other).expanduser().resolve()
        except:
            return False
        self = self.expanduser().resolve()
        return self.as_posix() == other.as_posix()


def ranID(*, length=12, size=1, only_alphanumeric=False):
    """Base64 encoded random ID.
    Args:
        length (int): Length of the ID string, rounded up to
            the closest multiple of 4. Default 12.
        size (int): Number of random IDs to generate.
            Default 1.
        only_alphanumeric (bool): If True, only alphanumeric
            characters are used. Default False.
    Returns:
        str: Random ID string.
    """
    if length < 1:
        raise ValueError("Length must be greater than 0!")
    if size < 1:
        raise ValueError("Size must be greater than 0!")
    if size > 1:
        return [ranID(length=length, only_alphanumeric=only_alphanumeric)
                for _ in range(size)]
    length = int(np.ceil(length/4))
    if only_alphanumeric:
        ran = ''
        for _ in range(length):
            while True:
                this_ran = ranID(length=4, only_alphanumeric=False)
                if this_ran.isalnum():
                    break
            ran += this_ran
        return ran
    else:
        random_bytes = os.urandom(3*length)
        return base64.urlsafe_b64encode(random_bytes).decode('utf-8')



def count_required_arguments(func):
    i = 0
    sig = inspect.signature(func)
    for param in sig.parameters.values():
        if (param.kind == inspect.Parameter.POSITIONAL_ONLY \
        or param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD) \
        and param.default == inspect.Parameter.empty:
            i += 1
    return i



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

# # How to use:
# # ===========

# # In master package __init__.py:
# from .aggregate_constants import merge_constants
# merge_constants( # Merge all providers into master namespace
#     entry_point_group="xtrack.constants",
#     target_module=__name__,
#     include_meta=True,      # also attach <group>_<plural>_meta
#     include_groups=True,    # also attach <group>_<plural>_groups (if providers defined any)
#     import_vars=False       # set to True if you also want the constants themselves at root
# )

# # In dependent package pyproject.toml:
# [project.entry-points."xtrack.constants"]
# particle_states = "xcoll.headers.particle_states:XcollParticleState"
# # you can list multiple tables:
# # particle_masses = "xcoll.headers.particle_masses:XcollParticleMass"



from __future__ import annotations

import sys
import numpy as np
from types import MappingProxyType
from typing import Dict, Tuple, Optional, Union

try:
    from importlib.metadata import entry_points  # Py3.10+
except ImportError:  # pragma: no cover
    from importlib_metadata import entry_points  # backport

from .constants import Constants

Number = Union[int, float]

class MergeError(ValueError): ...
class TableKey(tuple):
    # (group, noun); we also keep plural/reverse alongside the data store
    pass

def merge_constants(
    *,
    entry_point_group: str = "constants",
    target_module: Optional[str] = None,
    include_meta: bool = True,
    include_groups: bool = True,
    import_vars: bool = False,          # also re-export the individual constants
    overwrite_vars: bool = False,       # if True, allow overwriting same-valued vars
    order_by: str = "entrypoint",       # "entrypoint" | "alpha"
) -> Dict[Tuple[str, str], Dict[str, object]]:
    """
    Discover provider Constants via entry points and merge tables by (group, noun).

    Returns a dict keyed by (group, noun) with:
      {
        "plural": <plural>,
        "reverse": <reverse>,
        "map": Ordered[name -> number],
        "names": Optional[number -> name],   # only when reverse == "unique"
        "meta": Ordered[name -> {value, c_name, info}]   # when include_meta
        "groups": Ordered[group_name -> frozenset[number]]  # when include_groups
      }
    """
    # discover
    eps = entry_points()
    eps = eps.select(group=entry_point_group) if hasattr(eps, "select") else eps.get(entry_point_group, [])
    eplist = list(eps)
    if order_by == "alpha":
        eplist.sort(key=lambda e: (getattr(e, "name", ""), getattr(getattr(e, "dist", None), "name", "")))

    stores: Dict[Tuple[str, str], Dict[str, object]] = {}

    # track global collisions (in case providers didn’t share the same constants module)
    global_names: Dict[str, str] = {}            # NAME -> "dist/module"
    global_values: Dict[Tuple[str, str], Dict[Number, str]] = {}  # per (group, noun) when unique

    for ep in eplist:
        obj = ep.load()
        if not isinstance(obj, type) or not issubclass(obj, Constants):
            raise MergeError(f"Entry point {ep} does not load a Constants subclass (got {obj!r}).")

        g, n, p, r = obj.__group__, obj.__noun__, obj.__plural__, obj.__reverse__

        key = (g, n)
        store = stores.setdefault(key, {
            "plural": p,
            "reverse": r,
            "map": {},           # ordered
            "meta": {} if include_meta else None,
            "names": {} if r == "unique" else None,
            "groups": {} if include_groups and hasattr(obj, f"{p}_groups") else None,
        })

        # check compatibility
        if store["reverse"] != r:
            raise MergeError(f"Incompatible reverse modes for ({g},{n}): {store['reverse']!r} vs {r!r} from {ep}.")
        if store["plural"] != p:
            # We could normalize to first plural; better to require consistency
            raise MergeError(f"Inconsistent plural for ({g},{n}): {store['plural']!r} vs {p!r} from {ep}.")

        # provider maps
        prov_map: Dict[str, Number] = dict(getattr(obj, p))               # <group>_<plural>
        prov_meta: Optional[Dict[str, dict]] = dict(getattr(obj, f"{p}_meta")) if include_meta else None
        prov_groups: Optional[Dict[str, frozenset]] = None
        if include_groups and hasattr(obj, f"{p}_groups"):
            prov_groups = dict(getattr(obj, f"{p}_groups"))

        # name collisions (global)
        prov_source = f"{getattr(getattr(ep, 'dist', None), 'name', 'unknown')}/{obj.__module__}"
        for name in prov_map:
            if name in global_names and global_names[name] != prov_source:
                raise MergeError(f"Name {name!r} defined by both {global_names[name]} and {prov_source}.")
            global_names[name] = prov_source

        # merge forward map (ordered union)
        merged_map: Dict[str, Number] = store["map"]  # type: ignore
        for k, v in prov_map.items():
            if k in merged_map and merged_map[k] != v:
                raise MergeError(f"Different values for {k!r} in ({g},{n}): {merged_map[k]!r} vs {v!r} (from {prov_source}).")
            merged_map.setdefault(k, v)

        # reverse map (if unique)
        if r == "unique":
            merged_names: Dict[Number, str] = store["names"]  # type: ignore
            gv = global_values.setdefault(key, {})
            for k, v in prov_map.items():
                if v in merged_names and merged_names[v] != k:
                    raise MergeError(f"Value {v!r} used by {merged_names[v]!r} and {k!r} in ({g},{n}).")
                if v in gv and gv[v] != k:
                    raise MergeError(f"Global value collision {v!r} for {gv[v]!r} and {k!r} in ({g},{n}).")
                merged_names.setdefault(v, k)
                gv.setdefault(v, k)

        # meta merge (must not disagree)
        if include_meta and prov_meta is not None:
            merged_meta: Dict[str, dict] = store["meta"]  # type: ignore
            for k, md in prov_meta.items():
                if k in merged_meta:
                    # require same c_name and info if both present
                    prev = merged_meta[k]
                    if any(prev.get(f) != md.get(f) for f in ("value", "c_name", "info")):
                        raise MergeError(f"Metadata conflict for {k!r} in ({g},{n}).")
                merged_meta.setdefault(k, md)

        # groups merge (only if provider classes defined groups)
        if include_groups and hasattr(obj, f"{p}_groups"):
            prov_groups = dict(getattr(obj, f"{p}_groups"))  # name -> tuple
            merged_groups = store["groups"]  # name -> tuple
            for gname, gseq in prov_groups.items():
                if gname in merged_groups and merged_groups[gname] != gseq:
                    raise MergeError(f"Group {gname!r} differs between providers for ({g},{n}).")
                merged_groups.setdefault(gname, gseq)


        # optionally import individual variables into target (numbers/bools only)
        if import_vars and target_module:
            tgt = sys.modules[target_module]
            # NOTE: at master we re-export plain numbers/bools (no per-const docstrings)
            for k, v in prov_map.items():
                if hasattr(tgt, k) and not overwrite_vars:
                    raise MergeError(f"{target_module} already has {k}; set overwrite_vars=True to replace.")
                setattr(tgt, k, v)

    # write merged surfaces into target module
    if target_module:
        tgt = sys.modules[target_module]
        for (g, n), data in stores.items():
            p = data["plural"]
            # forward map
            setattr(tgt, f"{g}_{p}", MappingProxyType(dict(data["map"])))  # type: ignore
            # reverse (if unique)
            if data["reverse"] == "unique":
                setattr(tgt, f"{g}_{n}_names", MappingProxyType(dict(data["names"])))  # type: ignore
            # meta (optional)
            if include_meta and data["meta"] is not None:
                setattr(tgt, f"{g}_{p}_meta", MappingProxyType(dict(data["meta"])))  # type: ignore
            # groups (optional)
            if include_groups and data["groups"] is not None:
                setattr(tgt, f"{g}_{p}_groups", MappingProxyType(dict(data["groups"])))  # type: ignore
            # __all__ housekeeping (minimal)
            all_list = list(getattr(tgt, "__all__", []))
            for nm in (f"{g}_{p}",):
                if nm not in all_list: all_list.append(nm)
            if data["reverse"] == "unique":
                nm = f"{g}_{n}_names"
                if nm not in all_list: all_list.append(nm)
            if include_meta:
                nm = f"{g}_{p}_meta"
                if nm not in all_list: all_list.append(nm)
            if include_groups and data["groups"] is not None:
                nm = f"{g}_{p}_groups"
                if nm not in all_list: all_list.append(nm)
            setattr(tgt, "__all__", all_list)

    return stores



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xtrack as xt

from ..geometry import XcollGeometry
from ...materials import Material
from ...interaction_record import InteractionRecord
from ...general import _pkg_root


class EverestEngine(xt.BeamElement):
    _xofields = {}

    allow_track = False

    _depends_on = [Material, InteractionRecord, xt.RandomUniform, xt.RandomExponential,
                   xt.RandomNormal, xt.RandomRutherford, xt.Drift, XcollGeometry]

    _extra_c_sources = [
        _pkg_root.joinpath('scattering_routines','geometry','rotation.h'),
        _pkg_root.joinpath('scattering_routines','everest','constants.h'),
        _pkg_root.joinpath('scattering_routines','everest','everest.h'),
        _pkg_root.joinpath('scattering_routines','everest','properties.h'),
        _pkg_root.joinpath('scattering_routines','everest','ionisation_loss.h'),
        _pkg_root.joinpath('scattering_routines','everest','multiple_coulomb_scattering.h'),
        _pkg_root.joinpath('scattering_routines','everest','nuclear_interaction.h'),
        _pkg_root.joinpath('scattering_routines','everest','crystal_parameters.h'),
        _pkg_root.joinpath('scattering_routines','everest','amorphous.h'),
        _pkg_root.joinpath('scattering_routines','everest','jaw.h'),
        _pkg_root.joinpath('scattering_routines','everest','channelling.h')
    ]



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from warnings import warn
import scipy.constants as sc

import xtrack as xt


def prepare_rf_sweep(line, *, cavities=None, sweep=None, sweep_per_turn=None, num_turns=None):
    rf_sweep = RFSweep(line=line, cavities=cavities)
    rf_sweep.prepare(sweep=sweep, sweep_per_turn=sweep_per_turn, num_turns=num_turns)


class RFSweep:
    def __init__(self, line, cavities=None):
        self.line = line
        if line.env.ref_manager is None:
            raise ValueError("Environment must have a `ref_manager` to use RFSweep!" \
                           + "Do not use `optimize_for_tracking` as it will "
                           + "disable expressions, which are needed for RFSweep.")
        line.enable_time_dependent_vars = False  # To be able to twiss in the prepare step
        self.env = line.env
        self.cavities = cavities
        self.env['rf_sweep_df'] = 0
        self._prepared = False

    @property
    def current_sweep_value(self):
        return self.env['rf_sweep_df']

    def prepare(self, *, sweep=None, num_turns=None, sweep_per_turn=None):
        # TODO: if using sweep and num_turns, the sweep should stop after num_turns,
        #       but if using sweep_per_turn, it should continue indefinitely.
        if not self._prepared:
            if sweep_per_turn is not None:
                if np.isclose(sweep_per_turn, 0):
                    raise ValueError("Variable `sweep_per_turn` must be non-zero!")
                if sweep is not None:
                    raise ValueError("Provide either `sweep` or `sweep_per_turn`, not both.")
                if num_turns is not None:
                    raise ValueError("Variable `num_turns` cannot be set when using `sweep_per_turn`.")
                self.sweep_per_turn = sweep_per_turn
            elif sweep is not None:
                if np.isclose(sweep, 0):
                    raise ValueError("Variable `sweep` must be non-zero!")
                if num_turns is None:
                    num_turns = int(abs(sweep))
                if num_turns <= 0:
                    raise ValueError("When using `sweep`, `num_turns` must be a positive integer.")
                self.sweep_per_turn = sweep / num_turns
            else:
                raise ValueError("Either `sweep` or `sweep_per_turn` must be provided.")
            self._get_cavity_data()
            self._install_zeta_shift()
            self.tw = self.line.twiss()
            self.reset() # Initialize rf_sweep_df
            self.env['rf_sweep'].dzeta = f"{self.L} * rf_sweep_df / ({self.f_RF} + rf_sweep_df)"
            for cavs in self.cavities:
                for cav in cavs['names']:
                    scale_factor = int(np.round(cavs['freq'] / self.f_RF))
                    if scale_factor == 1:
                        self.env.ref[cav].frequency += self.env.ref["rf_sweep_df"]
                    else:
                        self.env.ref[cav].frequency += scale_factor * self.env.ref["rf_sweep_df"]
            self.line.enable_time_dependent_vars = True
            print("Enabled time-dependent variables in the line.")
            self._prepared = True


    def reset(self):
        if self.sweep_per_turn is None:
            raise ValueError("RFSweep not prepared. Call `prepare` first.")
        t_turn = self.tw.T_rev0   # To start sweeping from turn 0
        current_time = self.env['t_turn_s']
        if current_time == 0:
            self.env['rf_sweep_df'] = f"(t_turn_s + {t_turn}) / {t_turn} * {self.sweep_per_turn}"
        else:
            self.env['rf_sweep_df'] = f"(t_turn_s - {current_time} + {t_turn}) / {t_turn} * {self.sweep_per_turn}"


    def info(self, sweep=None, num_turns=None):
        if sweep is not None or num_turns is not None:
            if sweep is not None:
                warn("The `sweep` argument is deprecated.", FutureWarning)
            if num_turns is not None:
                warn("The `num_turns` argument is deprecated.", FutureWarning)
            self.prepare(sweep=sweep, num_turns=num_turns)
        if self.sweep_per_turn is None:
            raise ValueError("RFSweep not prepared. Call `prepare` first.")
        existing_sweep = self.current_sweep_value

        beta0 = self.line.particle_ref.beta0[0]
        E = self.line.particle_ref.energy0[0]
        q = self.line.particle_ref.q0
        h = self.f_RF * self.L / beta0 / sc.c
        eta = self.tw.slip_factor

        bucket_height = np.sqrt(abs(q*self.V*beta0**2 / (np.pi*h*eta*E) * (
                            2*np.cos(self.phi) + (2*self.phi-np.pi)*np.sin(self.phi)
                        )))
        delta_shift = -self.sweep_per_turn / self.f_RF / eta
        bucket_turns = bucket_height * 2 / abs(delta_shift)
        print(f"The current frequency is {self.f_RF + existing_sweep}Hz, adding "
            + f"{self.sweep_per_turn}Hz per turn.")
        print(f"This sweep will move the center of the bucket with \u0394\u03B4 = "
            + f"{delta_shift:.4} per turn.")
        print(f"The bucket height is {bucket_height:.4}, so this implies the sweep "
            + f"will shift one bucket every {round(bucket_turns, 2)} turns.")
        if self.tw.qs * bucket_turns < 3:
            print(f"Warning: This is a very fast sweep, moving ~1 bucket in "
                + f"~{round(self.tw.qs * bucket_turns, 2)} synchrotron oscillations "
                + f"(on average). If the bucket moves faster than a particle "
                + f"can follow, that particle will move out of the bucket and "
                + f"remain uncaptured.")


    # For backward compatibility
    def track(self, sweep=0, particles=None, num_turns=0, verbose=True, *args, **kwargs):
        warn("Using RFSweep.track() is deprecated. Please use RFSweep.prepare() "
           + "and then xt.Line.track().", FutureWarning)
        self.prepare(sweep=sweep, num_turns=num_turns)
        self.line.track(particles=particles, num_turns=num_turns, *args, **kwargs)


    def _get_cavity_data(self):
        # Cavity data not yet extracted
        tt = self.line.get_table()
        tt_c = tt.rows[[nn for nn, ttt in zip(tt.name, tt.element_type)
                        if 'Cavity' in ttt and 'CrabCavity' not in ttt]]
        mask = tt_c.parent_name == None  # Unsliced cavities
        if self.cavities is None:
            self.cavities = np.unique(list(tt_c.name[mask]) + list(tt_c.parent_name[~mask]))
        else:
            if isinstance(self.cavities, str):
                self.cavities = [self.cavities]
            for cav in self.cavities:
                if cav not in self.env.elements:
                    raise ValueError(f"Cavity `{cav}` not found in environment!")
        if len(self.cavities) == 0:
            raise ValueError("No cavities found in the line!")
        freq = np.array([self.env[cav].frequency for cav in self.cavities])
        volt = np.array([self.env[cav].voltage for cav in self.cavities])
        lag  = np.array([self.env[cav].lag for cav in self.cavities])
        s_pos = []
        name_first_in_line = []
        for cav in self.cavities:
            if cav in tt_c.name:
                s_pos.append(tt_c.rows[cav].s_start[0])
                name_first_in_line.append(cav)
            elif cav in tt_c.parent_name:
                s_pos.append(tt_c.rows[tt_c.parent_name == cav].s_start.min())
                name_first_in_line.append(tt_c.rows[tt_c.parent_name == cav].name[0])
            else:
                raise ValueError(f"Cavity `{cav}` not found in the line!")
        idx  = np.argsort(freq)
        boundaries = np.where(~np.isclose(freq[idx][1:], freq[idx][:-1],
                                        rtol=1e-8, atol=1e-12)
                            )[0] + 1
        groups = []
        for gidx in np.split(idx, boundaries):
            groups.append({
                'names': self.cavities[gidx],
                'freq': np.mean(freq[gidx]),
                'voltage': volt[gidx].sum(),
                's_pos': np.array(s_pos)[gidx],
                'lag': lag[gidx],
                'name_first_in_line': np.array(name_first_in_line)[gidx],
            })
        self.cavities = sorted(groups, key=lambda g: (g['voltage'], g['freq']), reverse=True)
        if np.isclose(self.cavities[0]['voltage'], 0):
            raise ValueError("No active cavity found!")
        self.f_RF = self.cavities[0]['freq']
        self.phi = np.deg2rad(self.cavities[0]['lag'].mean())
        self.V = self.cavities[0]['voltage']
        self.L = self.line.get_length()
        if len(self.cavities) > 1:
            print("Found multiple cavities with different frequencies:")
            for g in self.cavities:
                print(f"{g['freq']}Hz  at {g['voltage']}V: {g['names']}")
            print(r"The sweep will be performed with respect to the highest "
                + f"voltage cavity at {self.f_RF}Hz. The other cavities will "
                + f"be shifted accordingly.")


    def _install_zeta_shift(self):
        if 'rf_sweep' not in self.env.elements:
            idx = np.argsort(self.cavities[0]['s_pos'])
            first_cavity = self.cavities[0]['name_first_in_line'][idx[0]]
            self.env.elements['rf_sweep'] = xt.ZetaShift(dzeta=0)
            self.line.insert('rf_sweep', at=f'{first_cavity}@start')



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import Material
from .atoms import Carbon, Oxygen, Hydrogen, Boron, Nitrogen, Silicon, Aluminium, Iron, Nickel, Zinc
from .database import db, _manually_add_material_to_db


# General compounds
# =================

CarbonDioxide = Material(components=[Carbon, Oxygen],   n_atoms=[1,2], density=1.951e-3, state='gas', temperature=273.15, pressure=1)
DryIce        = Material(components=[Carbon, Oxygen],   n_atoms=[1,2], density=1.564,    state='solid')
Water         = Material(components=[Hydrogen, Oxygen], n_atoms=[2,1], density=1.0,      state='liquid')


# Compound definitions from FLUKA
# ===============================

BoronNitride   = Material(components=[Boron, Nitrogen],            n_atoms=[1, 1],             density=2.25, state='solid')
Silica         = Material(components=[Silicon, Oxygen],            n_atoms=[1, 2],             density=1.92, state='solid')
BoronTrioxide  = Material(components=[Boron, Oxygen],              n_atoms=[2, 3],             density=2.46, state='solid')
AluminiumOxide = Material(components=[Aluminium, Oxygen],          n_atoms=[2, 3],             density=3.98, state='solid')
NiZnFerrite    = Material(components=[Nickel, Zinc, Iron, Oxygen], n_atoms=[0.45, 0.55, 2, 4], density=5.2,  state='solid')


# Metadata for database
# =====================

_manually_add_material_to_db(CarbonDioxide,  'CarbonDioxide',  short_name='CO2')
_manually_add_material_to_db(DryIce,         'DryIce')
_manually_add_material_to_db(Water,          'Water',          short_name='H2O',  fluka_name='WATER')
_manually_add_material_to_db(BoronNitride,   'BoronNitride',                      fluka_name='hBNpure')
_manually_add_material_to_db(Silica,         'Silica',         short_name='Sand', fluka_name='SiO2')
_manually_add_material_to_db(BoronTrioxide,  'BoronTrioxide',                     fluka_name='B2O3')
_manually_add_material_to_db(AluminiumOxide, 'AluminiumOxide',                    fluka_name='Al2O3')
_manually_add_material_to_db(NiZnFerrite,    'NiZnFerrite',                       fluka_name='NIZNFER2')


# Clean up namespace
del Carbon, Oxygen, Hydrogen, Boron, Nitrogen, Silicon, Aluminium, Iron, Nickel, Zinc
del Material
del db
